@@ -118,6 +118,8 @@ static void IpcpSendTerminateAck(struct fsm *, u_char);
 static void IpcpDecodeConfig(struct fsm *, u_char *, u_char *, int,
                              struct fsm_decode *);
 
+extern struct libalias *la;
+
 static struct fsm_callbacks ipcp_Callbacks = {
   IpcpLayerUp,
   IpcpLayerDown,
@@ -690,7 +692,7 @@ ipcp_SetIPaddress(struct ipcp *ipcp, struct in_addr myaddr,
   if (bundle->ncp.cfg.sendpipe > 0 || bundle->ncp.cfg.recvpipe > 0) {
     ncprange_getsa(&myrange, &ssgw, &ssmask);
     ncpaddr_getsa(&hisncpaddr, &ssdst);
-    rt_Update(bundle, sadst, sagw, samask);
+    rt_Update(bundle, sadst, sagw, samask, NULL, NULL);
   }
 
   if (Enabled(bundle, OPT_SROUTES))
@@ -915,7 +917,7 @@ ipcp_InterfaceUp(struct ipcp *ipcp)
 
 #ifndef NONAT
   if (ipcp->fsm.bundle->NatEnabled)
-    PacketAliasSetAddress(ipcp->my_ip);
+    LibAliasSetAddress(la, ipcp->my_ip);
 #endif
 
   return 1;
